/*
PlaySol - automatic solitaire playing program for windows
Copyright (C) 2006 Robert Hart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "StdAfx.h"
#include "solengine.h"
#define MKPT(x,y) (((y+10) <<16) + (x+10))


SolEngine::SolEngine(void)
: solWnd(NULL)
{
}

SolEngine::~SolEngine(void)
{
}

void SolEngine::scangrid(void)
{
	int i;
	solWnd->GetWindowRect(&solWndRect);
	if(solWndRect.Width() > 0 && solWndRect.Height() > 0){
		pilelocx[0] = (solWndRect.Width()-7*CARDWIDTH)/8;
		pilelocy[0] = CARDHEIGHT+12-1; /* measure */
		for(i=NPILES;i<NPILES+NDISCARD+2;i++){
			pilelocy[i]=6-1;
		}
		pilelocx[PSTOCK] = pilelocx[0];
		pilelocx[PDECK] = pilelocx[0]*2+CARDWIDTH;
		
		for(i=1;i<NPILES;i++){
			pilelocx[i]=pilelocx[0] + i*(CARDWIDTH+pilelocx[0]-1);
			pilelocy[i]=pilelocy[0];
		}
		for(i=0;i<NDISCARD;i++){
			pilelocx[i+PDISCARD1]=pilelocx[i+NPILES-NDISCARD];
		}
	}
	for(i=0;i<NPILES+NDISCARD+2;i++){
		scancolumn(i);
	}
}

void SolEngine::scancolumn(int i)
{
	int x = pilelocx[i];
	int y = pilelocy[i];
	int n = 0;
	gamepiles[i].clearpile();
	while (iscard(x,y)){
		if(ishiddencard(x,y)){
			gamepiles[i].setcardhidden(n);
			y+=DYHIDDEN;
		} else {
			gamepiles[i].setcard(n,identifycard(x,y));
			y+=DYSHOWN;
		}
		n++;
	}
	if(n && gamepiles[i].getcard(n-1).rank==Card::unknown_rank)
		gamepiles[i].setcardhidden(n-1);

	if(i>=NPILES+1 && n==0){
		/* stock and discard pile sometimes moves slightly */
		Card c;
		c =identifycard(x+2,y+1);
		if(c.rank==Card::unknown_rank){
			c =identifycard(x+4,y+2);
			if(c.rank==Card::unknown_rank){
				c =identifycard(x+6,y+3);
			}
		}
		gamepiles[i].setcard(0,c);
	}
}

BOOL SolEngine::iscard(int x, int y)
{
	COLORREF pixel[3];
	solDC = solWnd->GetDC();
	pixel[0]=solDC->GetPixel(x+CARDWIDTH-6,y);
//	solDC->SetPixel(x+CARDWIDTH-6,y,0x000000ff);
	pixel[1]=solDC->GetPixel(x+CARDWIDTH-6,y+1);
//	solDC->SetPixel(x+CARDWIDTH-6,y+1,0x000000ff);
	pixel[2]=solDC->GetPixel(x+CARDWIDTH-6,y-1);
//	solDC->SetPixel(x+CARDWIDTH-6,y-1,0x000000ff);
	solWnd->ReleaseDC(solDC);
	if(pixel[0]==0 && pixel[1]!=0 && pixel[2]!=0)
		return true;
	return false;
}

BOOL SolEngine::ishiddencard(int x, int y)
{
	COLORREF pixel[3];
	solDC = solWnd->GetDC();
	pixel[0]=solDC->GetPixel(x+10,y);
	pixel[1]=solDC->GetPixel(x+10,y+1);
	pixel[2]=solDC->GetPixel(x+10,y+3);
	solWnd->ReleaseDC(solDC);
	if(pixel[0]==0 && pixel[1]!=0 && pixel[2]==0)
		return true;
	return false;

	return false;
}
Card SolEngine::identifycard(int x, int y)
{
	Card target(Card::unknown_rank, Card::unknown_suit);
	int i, j;
	int fp2=0, fp=0;
	COLORREF pixel;
	solDC = solWnd->GetDC();
	fp=0; mytext = "";
	for(i=0;i<8;i++){
		for(j=0;j<4;j++){
			fp=fp<<1;
			pixel = solDC->GetPixel(x+i+2,y+j*2+3);
			//solDC->SetPixel(x+i+2,y+j*2+3,0x00ff0000);
			switch (pixel) {
				case 0x00000000:	
					target.suit = Card::black;
					fp++;
					break;
				case 0x000000ff:
					target.suit = Card::red;
					fp++;
					break;
				case 0x00ffffff:
					break;
				case 0x0000ff00:
					break;
				default:
					break;
			}
		}
	}
	for(j=0;j<32;j++){
		fp2=fp2<<1;
		pixel = solDC->GetPixel(x+j+12,y+DYSHOWN-1);
		//solDC->SetPixel(x+j+12,y+DYSHOWN-1,0x00ff0000);
		switch (pixel) {
			case 0x00000000:	
				fp2++;
				break;
			case 0x000000ff:
				fp2++;
				break;
			case 0x00ffffff:
				break;
			case 0x0000ff00:
				break;
			default:
				break;
		}
	}
	target.fp  = fp;
	target.fp2 = fp2;
	switch (fp){
		case 79409:
		case 1270544:
		case 20328704: /* AS on stock */
			target.rank = Card::ace;
			break;
		case 71587190:
		case 1145395040: /* 2C on stock */
			target.rank = 2;
			break;
		case 72709972:
		case 1163359552:
			target.rank = 3;
			break;
		case 288515952:
		case 321287936: /* 4C */
			target.rank = 4;
			break;
		case 2002081104:
		case 125130069:
		case 1968526592: /* 5D stock */
			target.rank = 5;
			break;
		case 58021201:
		case 928339216: /* 6S on stock */
			target.rank = 6;
			break;
		case 71587684:
		case 1145402944: /* 7S on stock */
			target.rank = 7;
			break;
		case 108352886:
		case 1733646176: /* 8S stock */
			target.rank = 8;
			break;
		case 125060215:
			target.rank = 9;
			break;
		case 1996977223:
		case 1886864503: /*10H on stock*/
			target.rank = 10;
			break;
		case 292672:
		case 4682752: /* JC on stock */
		case -1879572480: /* JD on stock */
			target.rank = Card::jack;
			break;
		case 2000963440:
		case 1950643968: /* QH on stock */
			target.rank = Card::queen;
			break;
		case 2001888324:
		case 1965442112:
		//case 74928486: /* KS on discard */
			target.rank = Card::king;
			break;
		case 0:
			break;
		default:
			target.rank = Card::unknown_rank;
	}
	switch (fp2){
		case 535822337: /* 7D */
			if (target.rank == Card::queen) target.rank = 9;
		case   58851186: /* KD */
		case  538851136: /* QD ?*/ 
		
		//case 538851136: /* JD */
		case  267911168: /* 10D */
		//case  267911171: /* 5D */
		case     65534: /* 3D */
		
			if(target.isred()) target.suit = Card::diamonds;
			if(target.isblack()) target.suit = Card::spades;
			break;

		case 133169153: /* 9C */
			if(target.rank==Card::queen) target.rank = 9;
		case   66584576: /* 9C */
		case   65028095: /* KC */
		case  130056191: /* KC2 */
		case 1102919168: /* QC */
		case -2089128960: /* QC2 */
		case      16256: /* 2C */
		case       8128: /* 3C */
		case   66584577: /* 5C */
		//case  133169153: /* 8C */
		case   33292288: /* 6C */
			if(target.isblack()) target.suit = Card::clubs;
			break;
		case 267911171: /* 9S */
			if(target.rank==Card::queen) target.rank = 9;
		case   58775039: /* KS */
		case  543894536: /* QS */
		case 1514143712: /* JS */
		case  133955585: /* 6S */
		case      16352: /* 2S */
		case      32704: /* 2S */
			if(target.isblack()) target.suit = Card::spades;
			if(target.isred()) target.suit = Card::diamonds;
		break;

		case 2147418143: /* 9H */
			if(target.rank == Card::queen) target.rank = 9;
		case -904117701: /* JH */
		case 1073709071: /* 5H */
			if(target.isred()) target.suit = Card::hearts;
			break;
		case 0:
		case -262017: /* stipple? */
			break;
		default:
			//target.rank= Card::unknown_rank;
			break;
	}
	solWnd->ReleaseDC(solDC);
	return target;
}


int SolEngine::play(int n)
{
	int won;
	char buff[32];
	BOOL restart=false;
	won=gamelog[0]=stalled=0;
	
	for(int i=0;i<n;i++){
		strcpy_s(gamelog,LOGSIZE,"NEW\r\n");
		restart=false;
		do {
			if(!play()){
				strcat_s(gamelog,LOGSIZE,"WON?\r\n");
				won++;
				restart=true;
				stalled=0;
			} else if(stalled++>30){
				strcat_s(gamelog,LOGSIZE,"RESIGN\r\n");
				restart=true;
				stalled=0;
				if(i+1<n) startNew(solWnd);
			}
		} while (!restart);
							
	}
	sprintf_s(buff,32,"Played: %d Won: %d\r\n", n, won);
	strcat_s(gamelog,LOGSIZE,buff);
	return won;
}

void SolEngine::dismissdialog(CWnd* dWnd){
	if(AppActivate(dWnd)){
		keybd_event('Y',0x45,0,0);
		keybd_event('Y',0x45,KEYEVENTF_KEYUP,0);
	}
}

void SolEngine::startNew(CWnd* dWnd){
	if(AppActivate(dWnd)){
		keybd_event(VK_F2,0x45,0,0);
		keybd_event(VK_F2,0x45,KEYEVENTF_KEYUP,0);
	}
}

bool SolEngine::AppActivate(CWnd* dWnd)
{
  if (dWnd == NULL)
    return false;

  dWnd->SendMessage(WM_SYSCOMMAND, SC_HOTKEY, (LPARAM) dWnd->m_hWnd);
  dWnd->SendMessage(WM_SYSCOMMAND, SC_RESTORE, (LPARAM) dWnd->m_hWnd);
  
  dWnd->ShowWindow(SW_SHOW);
  dWnd->SetForegroundWindow();
  dWnd->SetFocus();

  return true;
}
int SolEngine::play(void)
{
	int i,j;
	Card c;
	int outlow=Card::ace;
	int outhigh=Card::king;
	int putout;
	CWnd* solWnd2;
	solWnd = CWnd::FindWindow("Solitaire","Solitaire");
	solWnd2 = CWnd::FindWindow(NULL,"Solitaire");
	if(solWnd!=solWnd2)
		dismissdialog(solWnd2);

	if(!solWnd) return 0;
	scangrid();
	remain=52;
	for (i=0;i<NDISCARD;i++){
		c = gamepiles[i+PDISCARD1].gettopcard();
		if(c.rank>=outlow)
			outlow=c.rank;
		if(c.rank<=outhigh)
			outhigh=c.rank;
		remain -=c.rank;
	}
	putout = outhigh + 2;
	if(putout<2) 
		putout=2;
	if(remain==0)
		return 0;
	for (i=0;i<NPILES;i++){
		remain-=gamepiles[i].mnshown;
	}
	if(remain==0){
		sendmove(MTAUTODISCARD,0,0,0,0);
		return 0;
	}
	for (i=0;i<NPILES+2;i++){
		c = gamepiles[i].gettopcard();
		if(c.rank>0 && c.rank<=putout && i<NPILES+2){
			sendmove(MTPUTOUT,i,gamepiles[i].mnhidden+gamepiles[i].mnshown-1,0,0);
		}

		c = gamepiles[i].getbottomcard();

		if(c.rank==Card::unknown_rank && i<NPILES && (gamepiles[i].mnhidden+gamepiles[i].mnshown)){
			sendmove(MTREVEAL,i,gamepiles[i].mnhidden,0,0);
		}

		for (j=0;j<NPILES;j++){
			if(gamepiles[j].cantake(c) && !(c.rank==Card::king && i<NPILES && gamepiles[i].mnhidden==0)){
				sendmove(MTMOVE,i,gamepiles[i].mnhidden,j,gamepiles[j].mnhidden+gamepiles[j].mnshown);
				return 1;
			}
		}
	}
	sendmove(MTREVEAL,PSTOCK,0,0,0);
	return 1;
}

int SolEngine::sendmove(mttype type, int sourcepile, int sourcecard, int destpile, int destcard)
{
	LPARAM src, dest;
	char buf[32];
	switch (type){
	case MTREVEAL:
		src = MKPT(pilelocx[sourcepile]+3,pilelocy[sourcepile]+gamepiles[sourcepile].dy(sourcecard)+3);
		solWnd->SendMessage(WM_LBUTTONDOWN,MK_LBUTTON,src);
		solWnd->SendMessage(WM_LBUTTONUP,NULL,src);
		sprintf_s(buf,32,"REV%i\r\n",sourcepile);
		strcat_s(gamelog,LOGSIZE,buf);
		if(sourcepile<7) stalled = 0;
		return 1;
		break;
	case MTPUTOUT:
		src = MKPT(pilelocx[sourcepile],pilelocy[sourcepile]+gamepiles[sourcepile].dy(sourcecard));
		solWnd->SendMessage(WM_LBUTTONDOWN,MK_LBUTTON,src);
		solWnd->SendMessage(WM_LBUTTONUP,NULL,src);
		solWnd->SendNotifyMessage(WM_LBUTTONDBLCLK,MK_LBUTTON,src);
		solWnd->SendNotifyMessage(WM_LBUTTONDOWN,MK_LBUTTON,src);
		solWnd->SendNotifyMessage(WM_LBUTTONUP,NULL,src);
		
		//solWnd->SendNotifyMessage(WM_LBUTTONDBLCLK,MK_LBUTTON,src);
		sprintf_s(buf,32,"PUT%i\r\n",sourcepile);
		strcat_s(gamelog,LOGSIZE,buf);
		return 1;
		break;
	case MTAUTODISCARD:
		src = MKPT(1,1);
		solWnd->SendNotifyMessage(WM_RBUTTONDOWN,MK_RBUTTON,src);
		solWnd->SendNotifyMessage(WM_RBUTTONUP,NULL,src);
		strcat_s(gamelog,LOGSIZE,"AUTO\r\n");
		return 1;
		break;
	case MTMOVE:
		src = MKPT(pilelocx[sourcepile],pilelocy[sourcepile]+gamepiles[sourcepile].dy(sourcecard));
		dest = MKPT(pilelocx[destpile],pilelocy[destpile]+gamepiles[destpile].dy(destcard));
		solWnd->SendMessage(WM_LBUTTONDOWN,MK_LBUTTON,src);
		solWnd->SendMessage(WM_MOUSEMOVE,MK_LBUTTON,dest);
		solWnd->SendMessage(WM_LBUTTONUP,NULL,dest);
		sprintf_s(buf,32,"MOV%i:%i %i:%i\r\n",sourcepile, sourcecard, destpile, destcard);
		strcat_s(gamelog,LOGSIZE,buf);
		stalled = 0;
		return 1;
		break;
	default:
		break;
	}
	return 0;
}

char *SolEngine::gettext(){
	int i;
	static char stext[LOGSIZE];
	strcpy_s( stext, LOGSIZE, "" );
	for (i=0;i<NPILES+NDISCARD+2;i++){
		strcat_s( stext, LOGSIZE, gamepiles[i].gettext());
		strcat_s( stext, LOGSIZE, "\r\n");
	}
	return stext;
}

char *SolEngine::getlog(){
	return gamelog;
}

Card::Card(){
	rank = unknown_rank;
	suit = unknown_suit;
}

Card::~Card(){

}

Card::Card(int newrank, Suit newsuit){
	rank = (Rank)newrank;
	suit = (Suit)newsuit;
}

BOOL Card::isred(){
	if(suit == Card::red || suit == Card::hearts || suit == Card::diamonds) return true;
	return false;
}

BOOL Card::isblack(){
	if(suit == Card::black || suit == Card::clubs || suit == Card::spades) return true;
	return false;
}

#define SIZE_OF_CARD_NAMES 3
char* Card::NameOf() {   // Get the name of a card. 
	static char szName[SIZE_OF_CARD_NAMES];
	static char *Numbers[] = {
		"?", "A", "2", "3", "4", "5", "6", "7", "8", "9",
		"X", "J", "Q", "K"
	};
	static char *Suits[] = {
		"R", "H", "D", "B", "C", "S"
	};


	if( rank > 0)
		strcpy_s( szName, SIZE_OF_CARD_NAMES, Numbers[rank] );
	else
		strcpy_s ( szName,SIZE_OF_CARD_NAMES,"?");
	if (suit >= 0)
		strcat_s( szName, SIZE_OF_CARD_NAMES, Suits[suit] );
	else
		strcat_s( szName, SIZE_OF_CARD_NAMES, "?" );
	return szName;
}

Pile::Pile(void){
	mnhidden = 0;
	mnshown = 0;
}

Pile::~Pile(){

}
BOOL Pile::isempty(void){
	if(mnhidden == 0 && mnshown ==0) {return true;} else {return false;}
}

int Pile::nhidden(void){
  return mnhidden;
}

int Pile::ncards(void){
  return mnshown + mnhidden;
}

int Pile::dy(int i){
	if(i==0) return 0;
	if(i<mnhidden) return i*DYHIDDEN;
	return mnhidden*DYHIDDEN+(i-mnhidden)*DYSHOWN;
}

Card Pile::getcard(int i){
	return cards[i];
}

void Pile::setcard(int i, Card c){
  cards[i] = c;
  if (mnhidden+mnshown-1<i){
    mnshown=i-mnhidden+1;
  }
}

void Pile::removecard(int i)
{
  if (mnhidden+mnshown-1>i){
    mnshown=i-mnhidden-1;
    if(mnshown<0) {
      mnhidden+=mnshown;
      mnshown = 0;
    }
  }
}

void Pile::setcardhidden(int i)
{
	if (mnhidden-1<i){
		if (mnhidden+mnshown-1<i){
			mnshown=0;
		} else {
			/* this could be crack */
			mnshown-=(i-mnhidden+1);					
		}
		mnhidden=i+1;
	}
}

void Pile::clearpile()
{
	mnhidden=mnshown=0;
}

Card Pile::gettopcard(void)
{
	return getcard(mnhidden+mnshown-1);		
}

Card Pile::getbottomcard(void)
{
	if(mnshown) return getcard(mnhidden);
	return Card();
}

BOOL Pile::cantake(Card c)
{
	if(isempty() && c.rank == Card::king) {
		return true;
	} 
	if(c.rank == gettopcard().rank-1 && ((gettopcard().isred() && c.isblack()) || (gettopcard().isblack() && c.isred()))){
		return true;
	} 
	return false;
}
#define SIZE_OF_PILE_TEXT SIZE_OF_CARD_NAMES*52
char *Pile::gettext()
{
	static char stext[SIZE_OF_PILE_TEXT];
	int i;
	strcpy_s (stext, SIZE_OF_PILE_TEXT, "");
	for(i=0;i<mnhidden;i++){
		strcat_s (stext, SIZE_OF_PILE_TEXT, ":");
	}
	for(i=mnhidden;i<mnhidden+mnshown;i++){
		strcat_s (stext, SIZE_OF_PILE_TEXT, cards[i].NameOf());
	}
	return stext;
}
