/*
PlaySol - automatic solitaire playing program for windows
Copyright (C) 2006 Robert Hart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#pragma once

class Card
{
public:
	enum Suit {
		unknown_suit = -1,
		red,
		hearts,
		diamonds,
		black,
		clubs,
		spades
	};
	enum Rank {
		unknown_rank = -1,
		ace = 1,
		jack = 11,
		queen = 12,
		king = 13
	};

	Card(void);
	Card(int rank, Suit suit);
	BOOL isred();
	BOOL isblack();
	~Card(void);
	Suit suit;
	int rank;
	int fp, fp2;
	char *NameOf();
};

class Pile
{
public:
	Pile(void);
	~Pile(void);
	Card gettopcard();
	Card getbottomcard();
	Card getcard(int i);
	void setcard(int i, Card c);
	void setcardhidden(int i);
	void removecard(int i);
	void clearpile();
	BOOL isempty();
	int nhidden();
	int ncards();
	int dy(int i);
	BOOL cantake(Card c);
	int mnhidden;
	int mnshown;
	char *gettext();
private:
	Card cards[52];
} ;

#define NPILES 7
#define NDISCARD 4
#define CARDWIDTH 71
/* check this */
#define CARDHEIGHT 96 
#define DYHIDDEN 3
#define DYSHOWN 15

class SolEngine
{
public:
	SolEngine(void);
	~SolEngine(void);
private:
	void scangrid(void);
	void scancolumn(int i);
	void dismissdialog(CWnd* dWnd);
	void startNew(CWnd* dWnd);
	bool AppActivate(CWnd* wnd);
	enum mttype {
		MTREVEAL,
		MTPUTOUT,
		MTAUTODISCARD,
		MTMOVE
	};
	enum piletype {
		PSTOCK=NPILES,
		PDECK,
		PDISCARD1,
		PDISCARD2,
		PDISCARD3,
		PDISCARD4,
	};

	int sendmove(mttype type, int sourcepile, int sourcecard, int destpile, int destcard);
	Card identifycard(int x, int y);
	BOOL iscard(int x, int y);
	BOOL ishiddencard(int x, int y);
	Pile gamepiles[NPILES+NDISCARD+2];
	CWnd* solWnd;
	CDC* solDC;
	CRect solWndRect;
	CString mytext;
	int pilelocx[NPILES+NDISCARD+2];
	int pilelocy[NPILES+NDISCARD+2];
#define LOGSIZE 81920
	char gamelog[LOGSIZE];
	int stalled;
	int remain;
public:
	int play(void);
	int play(int n);
	int play(int n, BOOL restart);
	char *gettext();
	char *getlog();

};

